//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

// $Id: elf2armc.c 417 2007-03-26 22:34:25Z claudio $

/*
 * elf2armc 3.0
 *  
 * Create .hp file with new executable format
 *
 * Usage: elf2armc prog.exe prog.hp
 * Writes ARMCODE object to prog.hp
 *
 * See further specifications for details.
 *
 */
 
 
#include <stdio.h>
#include <libelf.h>
#ifdef __GNUC__
#include <unistd.h>
#else
#include <io.h>
#endif
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>


#ifndef O_BINARY
#define O_BINARY 0
#endif


#define SAT_DOARMC 0x2c10


//! Internal structure of an elf file

typedef struct
{
    // File descritor
    int             fd;
    Elf32_Ehdr     *ehdr;
    Elf32_Shdr     *shdr;
    Elf            *elf;
    Elf_Scn        *scn;
    Elf_Data       *data;
} Elf_file;

typedef struct 
{
    unsigned int index;
    char name[20];
    unsigned int type;
    unsigned int address;
    unsigned int size;
    unsigned char entsize;
    unsigned char flags;
    unsigned char link;
    unsigned int info;
    unsigned char align;
    int offset;
} elf_Section_Table_Entry;


// *************************
// new elf2hp globals

int verbose=0;

// host CPU endian
int is_big_endian = 0;

int entry_point=-1;
int code_size=0;
int data_size=0;

#define SLACK_SYMBOL "__slack_space"
int slack_space=-1;

#define STARTUP_SYMBOL "__startup"
int startup_opt=-1;

#define STARTUP_OPTIONS_SYMBOL "__startup_options"


#define MEMREQ_SYMBOL "__memory_req"
int memory_requirements=-1;

#define ROMLIB_SYMBOL "__romlib_version"
#define DROMLIB_SYMBOL "__dromlib_version"

unsigned int romlib_version=0;

char *input_file=NULL;
char *output_file=NULL;

// ----------------------------------------
// ENDIAN AWARENESS
// < ibl ; 2006-10-30 >
// ----------------------------------------

typedef unsigned int USB32;

//! SWAP_B16_IN_B32 - swap 16 bits in 32 bits
/**
 *    dest    - pointer to where the swapped src wil be put
 *    src    - pointer to a 32 bit value to swap
 */
#define SWAP_B16_IN_B32(dest, src) (                    \
    *((USB32*)(dest)) =                        \
      (((*((USB32*)(src))) << 16) | ((*((USB32*)(src))) >> 16))    \
)

//! SWAP_B8_IN_B32 - swap 8 & 16 bits in 32 bits
/**
 *    dest    - pointer to where the swapped src wil be put
 *    src    - pointer to a 32 bit value to swap
 *
 * This macro will either switch to the opposite byte sex (Big Endian vs.
 * Little Endian) a 32 bit value.
 */
#define SWAP_B8_IN_B32(dest, src) (                    \
    SWAP_B16_IN_B32(dest, src),                    \
    (*((USB32*)(dest)) =                        \
      ((((*((USB32*)(dest))) & (USB32)0xff00ff00) >> 8) |        \
      (((*((USB32*)(dest))) & (USB32)0x00ff00ff) << 8)))        \
)

int
host_is_big_endian()
{
    unsigned int t = 0xcafebabe;
    
    return *((unsigned short *) &t) == 0xcafe;
}

unsigned int
change_byte_sex_uint32(unsigned int src)
{
    unsigned int trg;
    
    SWAP_B8_IN_B32(&trg,&src);
    return trg;
}

int
change_byte_sex_int32(int src)
{
    int trg;
    
    SWAP_B8_IN_B32(&trg,&src);
    return trg;
    
}


// PARSE ALL COMMAND LINE ARGUMENTS

int parse_arguments(char **argv, int argc)
{
char *string;

while(--argc) {
++argv;
string=*argv;
if(string[0]=='-') {
// PARSE OPTION

switch(string[1])
{
case 'v':
case 'V':
verbose=1;
break;
case 'h':
case '?':
printf("*** Elf2ARMC v3.0 ***\nThis program is part of HPGCC, (c) 2006 HPGCC Team\n\nUsage: elf2armc [options] input_file.exe [output_file.hp]\n\nOptions:\n" \
"-h -?  Help screen.\n" \
"-v -V  Verbose mode.\n" \
"-sn -Sn  Slack space override, set slack space to 2^n bytes, n=0 means no slack, n=1 is invalid.\n" \
"\n\n"
); 
return 0;
case 's':
case 'S':
slack_space=atoi(string+2);
if(slack_space>=12 || slack_space==1 || slack_space<0) {
fprintf(stderr,"Invalid slack space size. Use -h for help.\n");
return 0;
}
break;
default:
fprintf(stderr,"Invalid option. Use -h for help.\n");
return 0;

}

continue;
}

if(input_file==NULL) {
// GET INPUT FILE
input_file=(char *)malloc(strlen(string)+1);
if(!input_file) {
fprintf(stderr,"Memory allocation error. ** ABORTED **\n");
return 0;
}
strcpy(input_file,string);
continue;
}

if(output_file==NULL) {
// GET INPUT FILE
output_file=(char *)malloc(strlen(string)+10);
if(!output_file) {
fprintf(stderr,"Memory allocation error. ** ABORTED **\n");
return 0;
}
strcpy(output_file,string);
continue;
}

}

// FINISHED PROCESSING ALL ARGUMENTS

if(!input_file) {
fprintf(stderr,"No input file specified. Use -h for help.\n");
return 0;
}


if(!output_file) {
// copy input file with .hp extension

output_file=(char *)malloc(strlen(string)+10);
if(!output_file) {
fprintf(stderr,"Memory allocation error. ** ABORTED **\n");
return 0;
}

char *ptr=input_file+strlen(input_file)-1;
while(ptr>input_file && *ptr!='.') --ptr;
strncpy(output_file,input_file,ptr-input_file);
ptr=output_file+(long) (ptr-input_file); 
strcpy(ptr,".hp");
}
else {

// add .hp extension if no extension was given
char *ptr=output_file+strlen(output_file)-1;
while(ptr>output_file && *ptr!='.') --ptr;
if(ptr==output_file) {
strcat(output_file,".hp");
}



}

if(verbose) printf("Input file=%s\n",input_file);
if(verbose) printf("Output file=%s\n",output_file);
return 1;

}





void libelf_failure()
{
    fprintf(stderr, "libelf::%s\n", elf_errmsg(elf_errno()));
	if(input_file) free(input_file);
	if(output_file) free(output_file);
    exit(1);
}

#define elf2hp_failure _failure


void _failure(char *msg)
{
	fprintf(stderr,"elf2hp::%s\n",msg);
	if(input_file) free(input_file);
	if(output_file) free(output_file);
	exit(-1);
}

/* Given Elf header, Elf_Scn, and Elf32_Shdr 
 * obtain config parameters addresses
 */
 
 
void
get_config_parameters(Elf *elf)
 {
      Elf_Scn *scn;
	  Elf_Data *data;
      Elf32_Shdr *shdr;
      Elf32_Ehdr *ehdr;
	  char *name;
      data = 0;
      int count = 0;
	  long slack_addr=-1,memreq_addr=-1,romlib_addr=-1,dromlib_addr=-1, startup_addr=-1;
	  long startup_opt_addr=-1;

   if ((ehdr = elf32_getehdr(elf)) == 0)
   {
   _failure("Can't get ELF headers.\n");
    }

// GET ENTRY POINT FROM MAIN HEADER

   if(entry_point==-1) entry_point=ehdr->e_entry;

// CALCULATE CODE AND DATA SIZE

	 scn=NULL;
     while ((scn = elf_nextscn(elf, scn)) != NULL) {
        if ((shdr = elf32_getshdr (scn)) != 0) {
		
				name=elf_strptr(elf,ehdr->e_shstrndx , (size_t)shdr->sh_name);
				if(name)
				{ 
				if (!strcmp(name,".text"))  code_size+=(shdr->sh_size+3)&(~3);
				if (!strcmp(name,".rodata"))  code_size+=(shdr->sh_size+3)&(~3);
				if (!strcmp(name,".romglobals"))  data_size+=(shdr->sh_size+3)&(~3);
				if (!strcmp(name,".data"))  data_size+=(shdr->sh_size+3)&(~3);
				}
                
         }
     }

	
// GET SYMBOL TABLE SECTION  
	  
	 scn=NULL;
     while ((scn = elf_nextscn(elf, scn)) != NULL) {
        if ((shdr = elf32_getshdr (scn)) != 0) {
                if (shdr->sh_type == SHT_SYMTAB) {
                    break;
                }
         }
     }
     
	 if(scn==NULL) return;
	  
// FIND CONFIGURATION SYMBOLS	  
	  
      if ((data = elf_getdata(scn, NULL)) == 0 || data->d_size == 0)	return;
  
      Elf32_Sym *esym = (Elf32_Sym*) data->d_buf;
      Elf32_Sym *lastsym = (Elf32_Sym*) ((char*) data->d_buf + data->d_size);
     
        
      for (; esym < lastsym; esym++)
          if (esym->st_value) { 
              
                name = elf_strptr(elf,shdr->sh_link , (size_t)esym->st_name);
                if(!name)
                    libelf_failure();  
					
				
                if(!strcmp(name,SLACK_SYMBOL))
					slack_addr=esym->st_value;
                if(!strcmp(name,MEMREQ_SYMBOL)) 
					memreq_addr=esym->st_value;
                if(!strcmp(name,DROMLIB_SYMBOL)) 
					dromlib_addr=esym->st_value;
                if(!strcmp(name,ROMLIB_SYMBOL)) 
					romlib_addr=esym->st_value;
                if(!strcmp(name,STARTUP_SYMBOL)) 
					startup_addr=esym->st_value;
                if(!strcmp(name,STARTUP_OPTIONS_SYMBOL))
					startup_opt_addr=esym->st_value;
				
           
       }
	   
// FIND .config section
if(romlib_addr==-1) romlib_addr=dromlib_addr;

	  
	 scn=NULL;
     while ((scn = elf_nextscn(elf, scn)) != NULL) {
        if ((shdr = elf32_getshdr (scn)) != 0) {
		
				name=elf_strptr(elf,ehdr->e_shstrndx , (size_t)shdr->sh_name);
				if(name)
				{ 
				if (!strcmp(name,".config"))  break;
				}
                
         }
     }
	   
	 if(scn==NULL) return;
	  
// FIND CONFIGURATION SYMBOLS WITHIN .config SECTION  
	  
      if ((data = elf_getdata(scn, NULL)) == 0 || data->d_size == 0)	return;
 	   
	  if(slack_addr!=-1  && (slack_addr-shdr->sh_addr)<data->d_size) slack_space=*((int *) ( (char *)data->d_buf+(slack_addr-shdr->sh_addr)));
	  if(memreq_addr!=-1 && (memreq_addr-shdr->sh_addr)<data->d_size) memory_requirements=*((int *) ((char *)data->d_buf+(memreq_addr-shdr->sh_addr)));
	  if(romlib_addr!=-1 && (romlib_addr-shdr->sh_addr)<data->d_size) romlib_version=*((unsigned int *) ((char *)data->d_buf+(romlib_addr-shdr->sh_addr)));
	  if(startup_addr!=-1 && (startup_addr-shdr->sh_addr)<data->d_size) startup_opt=*((unsigned int *) ((char *)data->d_buf+(startup_addr-shdr->sh_addr)));


	  if(startup_opt!=-1) {
// FIND STARTUP SYMBOL IN .romglobals SECTION

	 scn=NULL;
     while ((scn = elf_nextscn(elf, scn)) != NULL) {
        if ((shdr = elf32_getshdr (scn)) != 0) {
		
				name=elf_strptr(elf,ehdr->e_shstrndx , (size_t)shdr->sh_name);
				if(name)
				{ 
				if (!strcmp(name,".romglobals"))  break;
				}
                
         }
     }
	   
	 if(scn==NULL) return;
	  
// FIND CONFIGURATION SYMBOL WITHIN .romglobals SECTION  
	  
      if ((data = elf_getdata(scn, NULL)) == 0 || data->d_size == 0)	return;
 	  
	  if(startup_opt_addr!=-1 && (startup_opt_addr-shdr->sh_addr)<data->d_size) *((unsigned int *) ((char *)data->d_buf+(startup_opt_addr-shdr->sh_addr)))=startup_opt;

	  }
	return;
}


/**
 *    Verify the Elf header, check if it is ARM ELF32 little endian\n
 *    If any problem, then bails out.
 */
int verify_header(Elf_file  *elf_file)
{
    //Check elf header:
    // See file libelf/elf_repl.h line 175
    
    // Open header
    if ((elf_file->ehdr = elf32_getehdr(elf_file->elf)) == NULL)
        libelf_failure();

    if (verbose)
    {
        printf( "\nVerifying ELF header:\n" );
        printf( "---------------------\n" );
        printf( "  Class:      " );
        if (elf_file->ehdr->e_ident[EI_CLASS]==ELFCLASS32)
            printf("Elf32\n");
        else if (elf_file->ehdr->e_ident[EI_CLASS]==ELFCLASS64)
            printf("Elf64\n");
        else if (elf_file->ehdr->e_ident[EI_CLASS]==ELFCLASSNONE)
            printf("None\n");
        else 
            printf("Unknown\n");

        printf("  Encoding:   ");
        if (elf_file->ehdr->e_ident[EI_DATA]==ELFDATA2LSB)
            printf("Little endian\n");
        else if (elf_file->ehdr->e_ident[EI_DATA]==ELFDATA2MSB)
            printf("Big endian\n");
        else if (elf_file->ehdr->e_ident[EI_DATA]==ELFDATANONE)
            printf("None\n");
        else 
            printf("Unknown\n");
    
        printf( "  ELFVersion: %s\n", ( EV_CURRENT ) ? "Current" : "Unknown");

        printf("  Type:       ");
        if (elf_file->ehdr->e_type==ET_EXEC)
            printf("EXEC\n");

        printf( "  Version:    0x%08X\n", elf_file->ehdr->e_version);
        printf( "  Machine:    0x%08X\n", elf_file->ehdr->e_machine);
        printf( "  Entry:      0x%08X\n", elf_file->ehdr->e_entry);
    }
    
    // Check class
    if (elf_file->ehdr->e_ident[EI_CLASS]!=ELFCLASS32)
        elf2hp_failure("Stoping: Bad class value");
    // Check encoding
    if (elf_file->ehdr->e_ident[EI_DATA]!=ELFDATA2LSB)
        elf2hp_failure("Stoping: Bad encoding value");
    // Check type
    if (elf_file->ehdr->e_type!=ET_EXEC)
        elf2hp_failure("Stoping: Bad type value");
    // Check machine compatibility
    if (elf_file->ehdr->e_machine!=0x28)
        elf2hp_failure("Machine type incompatible");
    
    return(1);
}


void dump_sections(FILE *out,Elf *elf)
{

      Elf_Scn *scn;
	  Elf_Data *data;
      Elf32_Shdr *shdr;
      Elf32_Ehdr *ehdr;
	  char *name;

// SCAN SECTIONS TO DUMP

   if ((ehdr = elf32_getehdr(elf)) == 0)
   {
   _failure("Can't get ELF headers.\n");
    }
	  
	 scn=NULL;
     while ((scn = elf_nextscn(elf, scn)) != NULL) {
        if ((shdr = elf32_getshdr (scn)) != 0) {
		
				name=elf_strptr(elf,ehdr->e_shstrndx , (size_t)shdr->sh_name);
				if(name)
				{ 
				if ( (!strcmp(name,".text")) ||
				     (!strcmp(name,".rodata")) ||
				     (!strcmp(name,".romglobals")) ||
					 (!strcmp(name,".data")) 
					 )
					 {
					    if ((data = elf_getdata(scn, NULL)) == 0 || data->d_size == 0) continue;
						char padding[4]={0,0,0,0};
						if(verbose) printf("Dump section %s, %d bytes\n" ,name,shdr->sh_size);
						fwrite(data->d_buf,shdr->sh_size,1,out);
						if(shdr->sh_size&3) fwrite(padding,4-(shdr->sh_size&3),1,out);
					 		 
					 
					 }
				}
                
         }
     }


}


int
main(int argc,char **argv)
{

	Elf_file elf_file;

	//verbose=1;
	// are we running on a big endian CPU ?
	is_big_endian = host_is_big_endian();
    
	if(!parse_arguments(argv,argc)) {
		if(input_file) free(input_file);
		if(output_file) free(output_file);
		return(-1);
	}

	if (elf_version(EV_CURRENT) ==EV_NONE)
        _failure("Libelf out of date.");

	if ((elf_file.fd = open(input_file,O_RDONLY | O_BINARY)) < 0)
		_failure("Can't open input file.\n");

	if ((elf_file.elf = elf_begin(elf_file.fd, ELF_C_READ, NULL)) == NULL)
		libelf_failure();

	verify_header(&elf_file);
 	get_config_parameters(elf_file.elf);

// SET DEFAULTS FOR CONFIG PARAMETERS NOT GIVEN IN EXECUTABLE

	if(romlib_version==0) {
		romlib_version=0X30303052; // "R000"
		// TEMPORARILY SWAP BYTE ORDER FOR CORRECT DISPLAY
		if(is_big_endian)
		    romlib_version = change_byte_sex_int32(romlib_version);
		
		fprintf(stderr,"** WARNING **: Using default romlib version R000\n");
	}
	else {
		// TEMPORARILY SWAP BYTE ORDER FOR CORRECT DISPLAY
		if(is_big_endian)
		    romlib_version = change_byte_sex_int32(romlib_version);
		
		if(verbose) printf("Romlib version override: %c%c%c%c\n",romlib_version&0xff,(romlib_version>>8)&0xff,(romlib_version>>16)&0xff,(romlib_version>>24)&0xff);
	}
	
	// RE-ADJUST BYTE ORDER FOR CONSISTENCY (WILL BE ALTERED LATER!)...
	if(is_big_endian)
	    romlib_version = change_byte_sex_int32(romlib_version);
	
	
	
	if(slack_space==-1) {
		slack_space=0;
		if(verbose) printf("Using default slack space = 0 bytes\n");
	}
	else {
		if(verbose) printf("Slack space override = 2^%d bytes\n",slack_space);
	}

	if(memory_requirements==-1) {
		memory_requirements=0X00070104;  // ALL RAM + 1KB STACK + 4KB TEMPOB
		if(verbose) printf("Using default memory requirements\n");
	}
	else {
		if(verbose) printf("Memory requirement override = %08X\n",memory_requirements);
	}
	if(entry_point==-1) { fprintf(stderr,"** WARNING **: No entry point found, assuming 0x1220000\n"); entry_point=0x01220000; }
	// WRITE .hp HEADER FILE ACCORDING TO PARAMETERS GIVEN

	struct {
	unsigned char hphp[8];
	unsigned char prologs [8];
	int romlib;
	int slack_offset;
	int memreq;
	int entrypt;
	} newexec;
    
	int progsize;

	strncpy((char *)newexec.hphp,"HPHP49-C",8);
	{
	    
	    // TODO:
	    // CAUTION:
	    // (BIG) ENDIAN HANDLING STILL REQUIRES SOME ATTENTION
	    // FIRST TESTS ON PPC64 WORK HOWEVER ...
	    // PROCEED *CAREFULLY* !!!
	    
	    // < ibl ; 2006-10-30 >
	    // Endian awareness added
	    
	    unsigned int sat_doarmc = SAT_DOARMC;
	    
	    // FAULTY BYTE SWAP REMOVED
	    // < ibl ; 2006-12-17 >
/*	    
	    if (is_big_endian)
		sat_doarmc = change_byte_sex_uint32(sat_doarmc);
*/
	    
	    newexec.prologs[0]=sat_doarmc&0xff;
	    newexec.prologs[1]=sat_doarmc>>8;
	}
	
	progsize=(code_size+data_size)*2+sizeof(newexec)*2-21+ ((slack_space)? 1<<(slack_space+1):0);
	
	if(progsize>=0x100000) {
        _failure("EXE is too big for calculator object.\n");	
	}

	if(verbose) printf("Total program size = %d\nCode object size=%d nib\n",code_size+data_size,progsize);
	    
	{
	    // ENDIAN CORRECTION 
	    // ALL VARIABLES ARE LOCALLY SHADOWED; NONE ALTERED
	    // < ibl ; 2006-10-30 >
	    
	    int _progsize,_romlib_version,_slack_space,_memory_requirements,_entry_point;
	    
	    // FAULTY BYTE SWAP REMOVED
	    // < ibl ; 2006-12-17 >
	    
	    // _progsize = is_big_endian ? change_byte_sex_int32(progsize) : progsize;
	    
	    _progsize = progsize;

	    newexec.prologs[2]=(_progsize<<4)&0xff;
	    newexec.prologs[3]=(_progsize>>4)&0xff;
	    newexec.prologs[4]=(_progsize>>12)&0xff;
	    
	    newexec.prologs[5]='C';
	    newexec.prologs[6]='C';
	    newexec.prologs[7]='3';

	    // FAULTY BYTE SWAP REMOVED
	    // < ibl ; 2006-12-17 >
	    
	    //_romlib_version = is_big_endian ? change_byte_sex_int32(romlib_version) : romlib_version;
	    
	    
	    _romlib_version = romlib_version;

	    newexec.romlib= _romlib_version;
	    
	    _slack_space = is_big_endian ? change_byte_sex_int32(slack_space) : slack_space;
	    newexec.slack_offset= _slack_space&0xff;
	    
	    _memory_requirements = is_big_endian ? change_byte_sex_int32(memory_requirements) : memory_requirements;
	    newexec.memreq= _memory_requirements;
	    
	    _entry_point = is_big_endian ? change_byte_sex_int32(entry_point) : entry_point;
	    newexec.entrypt= _entry_point;
	}
	
	
    FILE *out;
    
	
	
    if(!(out = fopen(output_file,"wb")))
        _failure("Can't open target file for writing.\n");
    
	fwrite(&newexec,sizeof(newexec),1,out);
	
    dump_sections(out,elf_file.elf);
	
	if(slack_space) {
	char *buffer=(char *)malloc(1<<slack_space);
	if(!buffer) _failure("Insufficient memory.\n");	
	memset(buffer,0,1<<slack_space);
	fwrite(buffer,1<<slack_space,1,out);
	free(buffer);
	}
    
    fclose(out);
    close(elf_file.fd);
	
	return 0;
}
