//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

// $Id: rompatch.c 626 2008-12-24 00:46:32Z claudio $



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Configure the rest conditional for GNU C (Unix/Linux,Cygwin) or MSVC

#ifdef __GNUC__
typedef long long LONGLONG;
typedef unsigned long long ULONGLONG;
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#define _fileno fileno
#else
typedef __int64 LONGLONG;
typedef unsigned __int64 ULONGLONG;
#include <io.h>
#endif


// Unix/Linux compatibility upgrade by Ingo Blank, 2006-09-25
// Tested on Novell SLED10

#ifdef __GNUC__

// POSIX, standard C don't know a _filelength() function
// < ibl ; 2006-09-25 >

off_t
_filelength(int fildes)
{
    struct stat sb;

    fstat(fildes,&sb);
    return sb.st_size;
}

#endif


unsigned int Add_CRC(unsigned int crc,unsigned int nibble)
{
   return ((((crc^nibble) & 15) * 4225)) ^(crc>>4);
}

#define ADDR_MASK ((unsigned long) ~3)

void sat_poke(unsigned char *bank,int sataddr,unsigned val,int nibbles)
{

    // GENERAL POKE ROUTINE, UP TO 8 NIBBLES

    LONGLONG *ptr=(LONGLONG *)((((unsigned long)bank)+(sataddr>>1)) & ADDR_MASK); // < ibl ; 2006-10-26 >

    ULONGLONG lval,mask2;
    unsigned long mask;
    int rot=(sataddr&7)<<2;

    lval=*ptr;
    mask=1ULL<<(nibbles<<2);
    mask--;
    val&=mask;
    mask2=((ULONGLONG)mask)<<rot;
    mask2=-mask2-1;
    *ptr=(((ULONGLONG)val)<<rot)|(lval&mask2);

}


unsigned sat_peek(unsigned char *bank,int sataddr,int nibbles)
{

    // GENERAL PEEK ROUTINE, UP TO 8 NIBBLES

    LONGLONG *ptr=(LONGLONG *)((((ULONGLONG)bank)+(sataddr>>1)) & ADDR_MASK); // < ibl ; 2006-10-26 >


    unsigned long mask;
    int rot=(sataddr&7)<<2;

    mask=1ULL<<(nibbles<<2);
    mask--;

return (unsigned)( (unsigned long)((*ptr)>>rot))&mask;
}
//@</func>

unsigned int pages[512], npages=0;

void findfreepages(unsigned char *bank, int baseaddr)
{
// ADD FREE PAGES TO THE TABLE
unsigned char *ptr=bank;
unsigned int addr;
int f;

for(f=0;f<0x3ff;++f,++ptr)
{
if(*ptr!=0xff) return;
}

addr=((unsigned int) baseaddr);

addr+=0x3ff;
addr&=~0x3ff;

pages[npages]=addr;
++npages;

}

#define SAT_DOLIB 0x2b40
#define SAT_DOBAK 0x2b62

// CONVERTED FROM ..\..\stub.bin

// LENGTH=116 NIBBLES

unsigned int StubObject[]= {
0x06F02B62,0x48270B00,0x43434750,0x494C2E33,0x2A2C0B42,0x48000410,0x43434750,0x494C2033,
0x52415242,0x202D2059,0x4E204F44,0x5020544F,0x45475255,0x00291121,0x00002E92
}; // END OF DATA



void insertlibrary(unsigned char *libdata,unsigned char *bank, int baseaddr)
{
// FIND FREE SPACE AFTER LIBRARIES
unsigned char *ptr=bank+0x1ffff;
unsigned long addr,endaddr;
int used,free,libsize;

if((sat_peek(libdata,0,5)!=SAT_DOLIB) && (sat_peek(libdata,0,5)!=SAT_DOBAK)) {
printf("Invalid library file\n");
return;
}
libsize=sat_peek(libdata,5,5);
libsize+=5+8+1;
libsize>>=1;

while(*ptr==0xff) --ptr;

++ptr;		// POINT TO FIRST 0xFF BYTE
addr=(unsigned long) (ptr-bank);
addr+=4;

used=addr-0x105;
free=0x20000-addr;

if(free<libsize) {
printf("Not enough space for library\n");
return;
}

// MAKE HOLE IN BANK
memmove(bank+0x105+free,bank+0x105,used);
memset(bank+0x105,0xff,free);

// INSTALL NEW LIBRARY
memcpy(bank+0x109,libdata,libsize-4);

addr=((free-1)<<9) | 0x53;
*((int *)(bank+0x105))=addr;

// CALCULATE HOW MANY PAGES TO ADD

endaddr=(0x105+free)&0xfffffc00;
addr=(0x105+libsize+0x3ff)&0xfffffc00 ;


while(addr<endaddr) {
//printf("Adding %0X\n",addr+baseaddr);
pages[npages]=addr+baseaddr;
++npages;
addr+=0x400;
}


}



unsigned int MMU_ADDRESS=0x3f000;
unsigned int ROMLIB_VERSION=0x3fffc;

int write_mmu_table(unsigned char *buffer)
{
int f;
unsigned int *ptr= (unsigned int *) (buffer+MMU_ADDRESS);
for(f=0;f<npages;++f)
{
	if(pages[f]==MMU_ADDRESS) break;
}
if(f==npages) {
	printf("Cannot create MMU table, incompatible ROM\n");
	return 0;
}

for(f=0;f<npages;++f)
{
if((pages[f]&(~0xfff))==MMU_ADDRESS) continue; // DON'T INCLUDE THE MMU TABLE PAGE
*ptr=pages[f]|0x2f;		// PAGES WITH READ-ONLY PERMISSION, CACHED AND BUFFERED
++ptr;
}



return 1;

}


void write_exe(unsigned char *buffer,const char *name)
{
int f;
FILE *han;
unsigned char *tmpbuffer[1024];
int length;

char *version=(char *)buffer+ROMLIB_VERSION;


han=fopen(name,"rb");
if(han==NULL) { printf("**ERROR**: Cannot open file \"%s\"\n",name); return; }


// PUT VERSION STAMP INTO MMU TABLE
length=fread(tmpbuffer,1,4,han);
if(length) memmove(version,tmpbuffer,4);

for(f=0;f<npages;++f)
{
if((pages[f]&(~0xfff))==MMU_ADDRESS) continue; // DON'T INCLUDE THE MMU TABLE PAGE
length=fread(tmpbuffer,1,1024,han);
if(length) memcpy(buffer+pages[f],tmpbuffer,length);
else break;
}
if(!feof(han))  printf("**ERROR**: Not enough space in ROM for the executable!!!\n");
printf("Space used by executable: %d pages\n",f);
fclose(han);

return;
}



int system_patches[]={

// USE ADDRESS=-1 TO MARK END-OF-PATCH

// ADDRESS , NUMPATCHES
   0X2C06  ,  1 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* =DIRECT_EVAL	*/ 				0X6713 			, 5 ,

// ADDRESS , NUMPATCHES
   0X2C15  ,  2 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* D=D+1.A	*/ 					0X7E  			, 2 ,
/* GOVLNG EVAL_ARMCODE */		0X06724D8		, 7 ,


// ADDRESS , NUMPATCHES
   0X2ea2  ,  1 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* =ROMP_EVAL_ADDR */			0X66FA 			, 5 ,

// ADDRESS , NUMPATCHES
   0X2F41  ,  1 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* =ROMP_EVAL_ADDR */	 		0X66FA 			, 5 ,



// ADDRESS , NUMPATCHES
   0X66C0  ,  15 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* LDR R2,[PC + #$8] */			0xe59f2008		, 8 ,
/* MOV R3,#$07F00000 */			0xE3A0367F		, 8 ,
/* STR R2,[R3 + #$4C] */		0xE583204C		, 8 ,
/* MOV PC,LR */					0xE1A0F00E		, 8 ,
/* PAGE VALUE TO LOAD */		0x0003F013		, 8 ,

// PC=$66E8
// INSTALL ROMLIB VIRTUAL MEMORY AREA
// SETUP_ROMLIB:

/* RSTK=C */					0X60			, 2 ,
/* LC(5) romlib_setup_addr */	0X066C043		, 7 ,
/* ARMSAT */					0XFFB08			, 5 ,
/* C=RSTK */					0X70			, 2 ,
/* RTN */						0X10			, 2 ,

//PC=$66FA
/* =ROMP_EVAL */				0x66ff			, 5 ,
//ROMP_EVAL:
/* A=DAT1.A */					0X341			, 3 ,
/* D1=D1+5 */					0X471			, 3 ,
/* D=D+1.A */					0X7E			, 2 ,
/* GOTO EVAL_ARMCODE */			0X01c6 			, 4 ,

//PC=$670b

// ADD MORE CODE HERE
// MAX. SPACE: PC < $670C


// ADDRESS , NUMPATCHES
   0X6713  ,  19 ,

// DIRECT EXECUTION OBJECT

//  MNEMONIC  					VALUE 			, NIBBLES
// DIRECT_EVAL:
/* C=DAT0  A */					0X641			, 3 ,
/* D0=D0-  5 */					0X481			, 3 ,
/* AD0EX */						0X231			, 3 ,
/* C=C+A   A */					0X2C			, 2 ,
/* D0=C */						0X431			, 3 ,
/* D0=D0+  5 */					0X461			, 3 ,
// HERE: A->OBJECT PROLOG, D0->NEXT OBJECT IN SECONDARY

// PC=$6724
//EVAL_ARMCODE:
/* GOSUB SAVEPTR */			    0X0737			, 4 ,
/* LC(8) _launch_legacy */		0X73			, 2 ,
								0X01300000		, 8 ,
//*ROMLIB_CALL
/* GOSUB SETUP_ROMLIB */		0XFB27			, 4 ,	// ***** WARNING: RELATIVE JUMP HERE ******
/* ARMSYS */					0XEFB08			, 5 ,
/* GOVLNG GETPTRLOOP */			0X05143D8		, 7 ,

//PC=$6742
//STANDARD_ROMP_EVAL:
/* GOSUB GETPTR */				0x08C7			, 4 ,
/* D1=D1-5 */					0x4C1			, 3 ,
/* DAT1=A.A	*/					0x141			, 3 ,
/* D=D-1.A */					0xFC			, 2 ,
/* LC(5) $02FEF */				0x02fef43		, 7 ,
/* A=C.A */						0xad			, 2 ,
/* PC=(A) */					0xc808			, 4 ,

//PC=$675B
// MAX. SPACE: PC < $675C


// ADDRESS , NUMPATCHES
   -1      ,  0



};

int mmu_patches[]={

// USE ADDRESS=-1 TO MARK END-OF-PATCH
// ADDRESS , NUMPATCHES
   0X66e0  ,  1 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* PAGE VALUE TO LOAD */		0x0003F013		, 8 ,

// ADDRESS , NUMPATCHES
   -1      ,  0

};

int system_patches2[]={

// USE ADDRESS=-1 TO MARK END-OF-PATCH

// ADDRESS , NUMPATCHES
   0xFFFFF  ,  1 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* =GOVLNG LIB_FIXSTO	*/ 		0XfffffD8		, 7 ,

// ADDRESS , NUMPATCHES
   -1      ,  0


};

int system_patches3[]={

// USE ADDRESS=-1 TO MARK END-OF-PATCH

// ADDRESS , NUMPATCHES
   0xFFFFF  ,  1 ,

//  MNEMONIC  					VALUE 			, NIBBLES
/* =GOVLNG PACKBANK_FIX	*/ 		0XfffffD8		, 7 ,


// ADDRESS , NUMPATCHES
   -1      ,  0


};


int system_fixsto[]={

	// ADDRESS , NUMPATCHES
   0xfffff  ,  5 ,

//  MNEMONIC  					VALUE 			, NIBBLES
//*LIB_FIXTO
/* LC(8) _launch_fixsto */		0X73			, 2 ,
								0X01300004		, 8 ,
//*ROMLIB_CALL_IRAMB
/* GOSBVL SETUP_ROMLIB */		0x066e8f8		, 7 ,
/* ARMSYS */					0XEFB08			, 5 ,
/* GOVLNG IRAMBUFF */			0X800F5D8		, 7 ,

// ADDRESS , NUMPATCHES
   -1      ,  0


};

int system_packbank[]={

	// ADDRESS , NUMPATCHES
   0xfffff  ,  5 ,

//  MNEMONIC  					VALUE 			, NIBBLES
//*LIB_FIXTO
/* LC(8) _launch_fixsto */		0X73			, 2 ,
								0X01300008		, 8 ,
//*ROMLIB_CALL_IRAMB
/* GOSBVL SETUP_ROMLIB */		0x066e8f8		, 7 ,
/* ARMSYS */					0XEFB08			, 5 ,
/* GOVLNG IRAMBUFF */			0X800F5D8		, 7 ,

// ADDRESS , NUMPATCHES
   -1      ,  0

};


unsigned char srch_fixsto[7]={0x61,0x15,0x85,0x83,0x5d,0x0f,0x80};
unsigned char srch_fixsto2[8]={0xda,0xd3,0xd7,0xa0,0xae,0x6d,0xfb,0xb3};

unsigned char srch_packbank[7]={0xf8,0x0c,0x67,0x80,0x5d,0x0f,0x80};



// APPLY PATCHES

void apply_patch(int *patches, unsigned char *bankptr)
{
int count;
int address;
int nnib, value;

while( (address=*patches++)!=-1) {
count=*patches++;

while(count--) {
value=*patches++;
nnib=*patches++;
sat_poke(bankptr,address,value,nnib);
address+=nnib;
}

}

}


// SEARCH FOR A STRING OF NIBBLES WITHIN A SPECIFIED REGION

int search_nibbles(unsigned char *srch,int size,int startsrch, int endsrch,unsigned char *bankptr)
{
int nibs;

for(;startsrch<endsrch;++startsrch)
{
for(nibs=0;nibs<size;nibs+=2)
{
if( ((unsigned char)sat_peek(bankptr,startsrch+nibs,2))!=srch[nibs>>1]) break;
}
if(nibs==size) return startsrch;
}

return -1;
}


int patch_system_bank(unsigned char *ptr)
{
int addr,addr2;
int f;

printf("Patching system bank...\n");

apply_patch(system_patches,ptr);


// CHECK FOR MMU TABLE PAGES

for(f=0;f<npages;++f)
{
	if(pages[f]==MMU_ADDRESS) break;
}
if(f==npages) {
	//PAGE NOT FOUND
	if(npages<1) {
	printf("Cannot create MMU table, incompatible ROM\n");
	return 0;
	}
	for(f=0;f<npages;++f)
	{
		if( (pages[f]&0xfff)==0 )
		{
			if((pages[f+1]==pages[f]+0x400) && (pages[f+2]==pages[f]+0x800) && (pages[f+3]==pages[f]+0xc00))
			{
				MMU_ADDRESS=pages[f];
				ROMLIB_VERSION=MMU_ADDRESS+0xffc;
				mmu_patches[2]=(MMU_ADDRESS&(~0xfff)) | 0x13;
				apply_patch(mmu_patches,ptr);
				break;
			}
		}
	}
	if(f==npages) {
	printf("Cannot create MMU table, incompatible ROM\n");
	return 0;
	}
}


// SEARCH FOR POSSIBLE MOVABLE PATCH POINTS

addr=search_nibbles(srch_fixsto,14,0x22000,0x26000,ptr);

if(addr==-1) {
printf("**ERROR**: FAILED TO PATCH FIXSTO - INCOMPATIBLE ROM IMAGE\n");
return 0;
}


addr2=search_nibbles(srch_fixsto2,16,addr,addr+0x100,ptr);

if(addr2==-1) {
printf("**ERROR**: FAILED TO PATCH FIXSTO - INCOMPATIBLE ROM IMAGE\n");
return 0;
}

// PREPARE PATCHES
system_patches2[0]=addr+7;
system_patches2[2]=((addr2-(addr+14)+0x800f5)<<8)|0xd8;

system_fixsto[0]=addr2;

apply_patch(system_patches2,ptr);
apply_patch(system_fixsto,ptr);


addr=search_nibbles(srch_packbank,14,0x22000,0x26000,ptr);

if(addr==-1) {
printf("**ERROR**: FAILED TO PATCH PACKBANK - INCOMPATIBLE ROM IMAGE\n");
return 0;
}

addr2=search_nibbles(srch_fixsto2,16,addr,addr+0x100,ptr);

if(addr2==-1) {
printf("**ERROR**: FAILED TO PATCH FIXSTO - INCOMPATIBLE ROM IMAGE\n");
return 0;
}

system_patches3[0]=addr+7;
system_patches3[2]=((addr2-(addr+28)+0x800f5)<<8)|0xd8;

system_packbank[0]=addr2;

apply_patch(system_patches3,ptr);
apply_patch(system_packbank,ptr);


return 1;

}

void patch_bank4(unsigned char *ptr)
{

	int k,h;
	char search[]="Extended 3";
	char replac[]="ARM/C Code";

	printf("Patching bank 4...");

	for(k=0;k<0x40000;++k)
	{
	for(h=0;(sat_peek(ptr,k+2*h,2)==search[h])&&(search[h]!=0);++h);
	if(search[h]==0) {
	// STRING FOUND AT CURRENT POINTER, REPLACE
		for(h=0;search[h]!=0;++h) sat_poke(ptr,k+2*h,replac[h],2);
		printf("Done\n");
	}
	}
	if(k==0x40000) printf("Not found\n");

}


void patch_bank5(unsigned char *ptr)
{

// REMOVED, MASD WON'T COMPILE BACK IF DECOMPILED AS ARMC

/*
unsigned char *pt2=ptr;
int k,h;
char search[]="EXT3";
char replac[]="ARMC";

printf("Patching bank 5...");

for(k=0;k<0x20000;++k,++pt2)
{
for(h=0;(pt2[h]==search[h])&&(search[h]!=0);++h);
if(search[h]==0) {
// STRING FOUND AT CURRENT POINTER, REPLACE
memcpy(pt2,replac,strlen(replac));
printf("Done\n");
break;
}
}
if(k==0x20000) printf("Not found\n");
*/
}





int main(int argc, char* argv[])
{
FILE *han;
unsigned char *buffer=(unsigned char *)malloc(3000000L);
unsigned char *ptr;
int offset=0,length,crcoff,totalcrc,f,banktype,banknum,crc;
int totalrombanks,lastvalid=-1;
int valid;

if(!buffer) {
	printf("Insufficient memory to run\n");
	return 0;
}


if(argc<2) {
	printf("Use:\n rompatch rom_binary.bin [romlib.bin] [library.lib]\n");
	free(buffer);
	return 0;
}

han=fopen(argv[1],"rb");
if(han==NULL) { printf("Cannot open file \"%s\"\n",argv[1]); exit(0); }
length=fread((void *)buffer,1,2*1024*1024,han);
fclose(han);
totalrombanks=(length+0x4000);
if(totalrombanks&0x1ffff) {
printf("Invalid ROM image %s\n",argv[1]);
exit(0);
}
totalrombanks>>=17;				// CALCULATE TOTAL NUMBER OF BANKS TO DISTINGUISH BETWEEN v92 AND PREVIOUS ONES
if(totalrombanks!=10 && totalrombanks!=11) {
printf("Invalid ROM image %s\n",argv[1]);
exit(0);
}
memmove(buffer+0x4000,buffer,length);
memset(buffer,0,0x4000);
offset=0x4000;
while(offset<length+0x4000) {
ptr=buffer+offset;
//printf("Scanning address %08X\n",offset);
banktype=sat_peek(ptr,0x200,2);
banknum=sat_peek(ptr,0x202,2);

if((banktype==0x32) || (banktype==0)) crcoff=sat_peek(ptr,0x100,5);
else crcoff=sat_peek(ptr,0x20a,5);

crcoff+=0x20a;

crc=sat_peek(ptr,crcoff,4);

if((banknum>15) || (banknum<=lastvalid) || ((banktype!=0x86) && (banktype!=0xf0) && (banktype!=0x0f) && (banktype!=0x32))) valid=0;
else {
totalcrc=0;
valid=0;

if((banktype==0xf || banktype==0x32 )) {

	if(offset>=0x40000) {
for(f=0x20a;f<crcoff;++f)
{
totalcrc=Add_CRC(totalcrc,sat_peek(ptr,f,1));
}
if(totalcrc==crc) valid=1;
	}
}
else {
	if((banktype==0xf0) && (offset<0x140000)) valid=0;		// reject user banks below user area
	else valid=1;
}
if(valid) { if(banktype!=0x86) lastvalid=banknum; }
}

//printf("CRC=%04X\n",crc);
if(valid)
printf("Saturn Bank %02X, type %02X, CRC=%04X\n",banknum,banktype,crc);
//else printf("Non-saturn bank at address %08X\n",offset);
//printf("crcoff=%08x\n",crcoff);


if(valid && ((banktype==0x32) && (banknum==1))) {
patch_system_bank(ptr);
}

if(valid &&(banknum==4)) {
patch_bank4(ptr);

}
if(valid &&(banknum==5)) {
patch_bank5(ptr);

}

if( (!valid) || (banktype!=0xf0)) // DON'T TAKE SPACE OUT OF USER BANKS
{
findfreepages(ptr,offset);
}

if(valid) {

totalcrc=0;
for(f=0x20a;f<crcoff;++f)
{
totalcrc=Add_CRC(totalcrc,sat_peek(ptr,f,1));
}
if(banknum>0) printf("Computed CRC=%08X\n",totalcrc);

if(banktype==0xf || banktype==0x32) {
	if(totalcrc!=crc) {
		printf("Bad CRC detected, fixing...\n");
		sat_poke(ptr,crcoff,totalcrc,4);
	}

}
}



	offset+=0x400;
}

// APPEND FRESH BANK
lastvalid++;

if((offset<0x140000) || ( (offset&0x1ffff)!=0))
{
	printf("Unable to add flash banks\n");
	free(buffer);
	exit(0);
}

memset(buffer+offset,0xff,0x20000);	// ADD NEW FLASH BANK
memset(buffer+offset+0x100,0,5);
buffer[offset+0x100]=0xf0;
buffer[offset+0x101]=lastvalid;
length+=0x20000;

ptr=buffer+offset;

{


// INSERT CUSTOM LIBRARY AND RESERVE REST OF THE BANK
char *libname;
unsigned char *library;
int len;

if(argc>=4) libname=argv[3];
else libname="satromlib.lib";

han=fopen(libname,"rb");
if(han==NULL) { printf("Cannot open file %s\n",libname); 	free(buffer); exit(0); }
len=_filelength(_fileno(han));
library=(unsigned char *)malloc(len);
if(!library) {
printf("Insufficient memory\n");
free(buffer);
exit(0);
}
if(fread(library,1,len,han)!=len) printf("**ERROR**: Failed to read library %s\n",libname);

fclose(han);

// MOVE OTHER LIBRARIES AND INSERT NEW LIBRARY

insertlibrary(library+8,ptr,offset);

}


offset+=0x20000;
ptr=buffer+offset;

// APPEND FRESH BANK
lastvalid++;

if((offset<0x140000) || ( (offset&0x1ffff)!=0))
{
	printf("Unable to add flash bank\n");
	free(buffer);
	exit(0);
}

memset(buffer+offset,0xff,0x20000);	// ADD NEW FLASH BANK
memset(buffer+offset+0x100,0,5);
buffer[offset+0x100]=0xf0;
buffer[offset+0x101]=lastvalid;
length+=0x20000;



{
// INSERT CUSTOM LIBRARY AND RESERVE REST OF THE BANK
// MOVE OTHER LIBRARIES AND INSERT NEW LIBRARY

insertlibrary((unsigned char *)StubObject,ptr,offset);

}

printf("Total recovered space: %d pages = %d kbytes\n",npages,npages*1024);
{
char *name;

if(!write_mmu_table(buffer)) {
	printf("Incompatible rom image or HPGCC3 already installed.\nPlease provide an original rom image.\n");

}
else {
	char romlib_version[5];


if(argc>=3) name=argv[2];
else name="romlib.bin";

write_exe(buffer, name);

	// EXTRACT ROMLIB VERSION NUMBER
	memmove(romlib_version,buffer+ROMLIB_VERSION,4);
	romlib_version[4]=0;



memmove(buffer, buffer+0x4000,length);

name="hpgccrom.bin";
han=fopen(name,"wb");
if(han==NULL) { printf("Cannot open file %s\n",name); exit(0); }
else printf("Writing patched rom to %s...",name);
if(fwrite(buffer,1,length,han)==length) {
	printf("Done\n" );
	printf("HPGCC3 installed, ROM revision %s.\n",romlib_version);

}
else printf("**ERROR**: Failed to write %s\n",name);

fclose(han);

}

}

free(buffer);

}

